
'''
This file is automatically generated; Do not edit it. :)
'''
VERSION_INFO = {
    'final': True,
    'version': '36.0.0rc5',
    'branch_nick': 'master',
    'revision_id': 'e2324c295e8de8d6f6b9b77fb21e6c0111077363',
    'revno': 13332
}
